import InterceptRule from './InterceptRule.js';
import AbortError from './lib/AbortError.js';
/**
 * Generate the mocked fetch with given context.
 * Multiple arrow functions are used to bypass `.bind(this)`
 * which will be frequently used if not doing so.
 */
export default class Fetcher {
    constructor(context) {
        Object.defineProperty(this, "options", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: {
                defaultRule: new InterceptRule('').reply((req) => {
                    throw new Error(`No onfetch rule matches this request to '${req.url}'`);
                }),
                AbortError,
                bypassRedirect: false,
            }
        });
        Object.defineProperty(this, "rules", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "context", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "original", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "mocked", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (...argArray) => {
                const request = new Request(...argArray);
                const { signal } = request;
                if (signal === null || signal === void 0 ? void 0 : signal.aborted) {
                    throw new this.options.AbortError();
                }
                let redirected = false;
                const respondPromise = (this.rules.find((rule) => rule.test(request)) || this.options.defaultRule)
                    .apply(request, {
                    original: this.original,
                    mocked: this.mocked,
                })
                    .then((response) => {
                    if (this.options.bypassRedirect) {
                        return response;
                    }
                    /**
                     * @see [A redirect status | Fetch Standard]{@link https://fetch.spec.whatwg.org/#redirect-status}
                     */
                    if (![301, 302, 303, 307, 308].includes(response.status)) {
                        return response;
                    }
                    if (request.redirect !== 'follow') {
                        throw new TypeError('Redirect mode not supported');
                    }
                    const locationString = response.headers.get('location');
                    if (locationString === null) {
                        return response;
                    }
                    const location = new URL(locationString, request.url);
                    if (!location.hash)
                        location.hash = new URL(request.url).hash;
                    redirected = true;
                    return this.mocked(new Request(location.href, request));
                })
                    .then((response) => {
                    if (redirected && !response.redirected) {
                        Object.defineProperty(response, 'redirected', {
                            value: true,
                        });
                    }
                    return response;
                });
                if (signal) {
                    return Promise.race([
                        respondPromise,
                        new Promise((resolve, reject) => {
                            signal.addEventListener('abort', () => {
                                reject(new this.options.AbortError());
                            });
                        }),
                    ]);
                }
                return respondPromise;
            }
        });
        /**
         * Adopt another context. Preserve active status.
         */
        Object.defineProperty(this, "adopt", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (context) => {
                if (this.context === context)
                    return;
                const beActive = this.isActive();
                this.restore();
                this.context = context;
                this.original = context.fetch;
                if (beActive)
                    this.activate();
            }
        });
        Object.defineProperty(this, "remove", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (rule) => {
                const ruleIndex = this.rules.indexOf(rule);
                if (ruleIndex === -1)
                    return false;
                this.rules.splice(ruleIndex, 1);
                return true;
            }
        });
        Object.defineProperty(this, "cleanAll", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => {
                this.rules.length = 0;
            }
        });
        Object.defineProperty(this, "addRule", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (input, init = {}) => {
                const rule = new InterceptRule(input, init);
                this.rules.push(rule);
                return rule;
            }
        });
        Object.defineProperty(this, "hasActive", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => this.rules.some((rule) => rule.isActive())
        });
        Object.defineProperty(this, "isActive", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => this.context.fetch === this.mocked
        });
        Object.defineProperty(this, "activate", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => {
                this.context.fetch = this.mocked;
            }
        });
        Object.defineProperty(this, "restore", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => {
                this.context.fetch = this.original;
            }
        });
        Object.defineProperty(this, "config", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (config) => {
                Object.assign(this.options, config);
            }
        });
        this.context = context;
        this.original = context.fetch;
    }
}
//# sourceMappingURL=Fetcher.js.map