import toCloneable from './toCloneable.js';
import { MessageProcessor, } from './Message.js';
/**
 * A fetch receiver
 * - For received requests, message to service worker.
 * - For messaged requests, send to the receiver. <- Intercept here.
 */
export default class Channel extends MessageProcessor {
    constructor(workerContainer) {
        super();
        Object.defineProperty(this, "port", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "fulfillList", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "beActive", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /**
         * For received requests, message to the service worker.
         */
        Object.defineProperty(this, "fetch", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (...args) => {
                const { fulfillList } = this;
                const port = await this.port;
                port.postMessage({
                    request: await toCloneable(new Request(...args)),
                    index: fulfillList.length,
                });
                return new Promise((resolve, reject) => {
                    fulfillList.push([resolve, reject]);
                });
            }
        });
        // Resolve when the service worker responds.
        Object.defineProperty(this, "statusResolve", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        this.registerContainer(workerContainer);
    }
    registerContainer(workerContainer) {
        this.beActive = false;
        this.fulfillList = [];
        this.port = this.preparePort(workerContainer)
            .finally(() => {
            workerContainer.addEventListener('controllerchange', () => {
                this.registerContainer(workerContainer);
            }, { once: true });
        });
    }
    async preparePort(workerContainer) {
        const { controller } = workerContainer;
        if (!controller) {
            return new Promise((resolve) => {
                workerContainer.addEventListener('controllerchange', () => {
                    resolve(this.preparePort(workerContainer));
                }, { once: true });
            });
        }
        const { port1, port2 } = new MessageChannel();
        port1.onmessage = this.onMessage;
        controller.postMessage({ onfetch: true }, [port2]);
        return port1;
    }
    /**
     * For messaged requests, send to the receiver.
     */
    async onRequestMessage(event) {
        const { data: { request, index } } = event;
        const port = await this.port;
        const responseOrError = await this.fetch(request.url, request)
            .then((response) => toCloneable(response))
            .catch((err) => err);
        port.postMessage({
            response: responseOrError,
            index,
        });
    }
    /**
     * For messaged responses, treat as a previous received request's response.
     */
    onResponseMessage(event) {
        const { data: { response, index } } = event;
        const { fulfillList } = this;
        if (!fulfillList) {
            throw new Error('Service worker not ready yet');
        }
        const fulfill = fulfillList[index];
        if (!fulfill) {
            throw new Error('Response came for unknown request');
        }
        if (response instanceof Error) {
            fulfill[1](response);
        }
        else {
            fulfill[0](new Response(response.body, response));
        }
        fulfillList[index] = null;
    }
    async onStatusMessage(event) {
        const { data: { status } } = event;
        const port = await this.port;
        this.beActive = status === 'on';
        const { statusResolve } = this;
        if (statusResolve) {
            statusResolve();
            this.statusResolve = null;
        }
        else {
            port.postMessage({ status });
        }
    }
    isActive() {
        return this.beActive;
    }
    async switchToStatus(status) {
        const port = await this.port;
        await new Promise((resolve) => {
            this.statusResolve = resolve;
            port.postMessage({ status });
        });
    }
}
//# sourceMappingURL=Channel.js.map