import toCloneable from './toCloneable.js';
import { MessageProcessor, } from './Message.js';
/**
 * A fetch forwarder
 * - For captured requests, message back to client.
 * - For messaged requests, message back the real response.
 */
export default class Worker extends MessageProcessor {
    constructor(scope) {
        super();
        Object.defineProperty(this, "scope", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "portMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "fulfillListMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "statusResolveMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        this.scope = scope;
        // Service worker listeners can only be added on the initial evaluation.
        scope.addEventListener('fetch', this.onFetch.bind(this));
        // For clients' port registration.
        scope.addEventListener('message', (event) => {
            if (typeof event.data !== 'object' || !('onfetch' in event.data))
                return;
            const { source } = event;
            if (!(source instanceof Client))
                return;
            if (this.portMap.has(source.id))
                return;
            const [port] = event.ports;
            this.portMap.set(source.id, port);
            port.onmessage = this.onMessage;
        });
    }
    /**
     * For captured requests, message back to the client.
     */
    onFetch(event) {
        const port = this.portMap.get(event.clientId);
        if (!port)
            return;
        const fulfillList = this.fulfillListMap.get(port);
        if (!fulfillList)
            return;
        event.respondWith((async () => {
            port.postMessage({
                request: await toCloneable(event.request),
                index: fulfillList.length,
            });
            return new Promise((resolve, reject) => {
                fulfillList.push([resolve, reject]);
            });
        })());
    }
    /**
     * For messaged requests, respond with real responses.
     */
    async onRequestMessage(event) {
        const { target, data: { request, index } } = event;
        if (!(target instanceof MessagePort) || !this.fulfillListMap.has(target)) {
            throw new Error('Request came from unrecognized source');
        }
        const responseOrError = await this.scope.fetch(request.url, request)
            .then((response) => toCloneable(response))
            .catch((err) => err);
        target.postMessage({
            response: responseOrError,
            index,
        });
    }
    /**
     * For messaged responses, treat as a previous captured request's response.
     */
    onResponseMessage(event) {
        const { target, data: { response, index } } = event;
        if (!(target instanceof MessagePort)) {
            throw new Error('Response came from unknown source');
        }
        const fulfillList = this.fulfillListMap.get(target);
        if (!fulfillList) {
            throw new Error('Response came from unregistered source');
        }
        const fulfill = fulfillList[index];
        if (!fulfill) {
            throw new Error('Response came for unknown request');
        }
        if (response instanceof Error) {
            fulfill[1](response);
        }
        else {
            // Client response may be of type 'error'.
            const nativeResponse = response.type === 'error'
                ? Response.error()
                : new Response(response.body, response);
            fulfill[0](nativeResponse);
        }
        fulfillList[index] = null;
    }
    onStatusMessage(event) {
        const { target, data: { status } } = event;
        if (!(target instanceof MessagePort)) {
            throw new Error('Status came from unknown source');
        }
        if (status === 'on') {
            this.fulfillListMap.set(target, []);
        }
        else {
            this.fulfillListMap.delete(target);
        }
        const statusResolve = this.statusResolveMap.get(target);
        if (statusResolve) {
            statusResolve();
            this.statusResolveMap.delete(target);
        }
        else {
            target.postMessage({ status });
        }
    }
    isActive() {
        return this.fulfillListMap.size > 0;
    }
    async switchToStatus(status) {
        const switchPromiseList = [...this.portMap.values()]
            .map((port) => new Promise((resolve) => {
            this.statusResolveMap.set(port, resolve);
            port.postMessage({ status });
        }));
        await Promise.all(switchPromiseList);
    }
}
//# sourceMappingURL=Worker.js.map